@@ -83,7 +83,7 @@ static xmlNode *firehose_response_parse(const void *buf, size_t len, int *error)
 	return node;
 }
 
-static int firehose_generic_parser(xmlNode *node, void *data)
+static int firehose_generic_parser(xmlNode *node, void *data __unused, bool *rawmode)
 {
 	xmlChar *value;
 	int ret = -EINVAL;
@@ -103,18 +103,27 @@ static int firehose_generic_parser(xmlNode *node, void *data)
 
 	xmlFree(value);
 
+	value = xmlGetProp(node, (xmlChar *)"rawmode");
+	if (value) {
+		if (xmlStrcmp(value, (xmlChar *)"true") == 0)
+			*rawmode = true;
+		xmlFree(value);
+	}
+
 	return ret;
 }
 
 static int firehose_read(struct qdl_device *qdl, int timeout_ms,
-			 int (*response_parser)(xmlNode *node, void *data),
+			 int (*response_parser)(xmlNode *node, void *data, bool *rawmode),
 			 void *data)
 {
 	char buf[4096];
 	xmlNode *node;
 	int error;
+	int resp = -EIO;
 	int ret = -EAGAIN;
 	int n;
+	bool rawmode = false;
 	struct timeval timeout;
 	struct timeval now;
 	struct timeval delta = { .tv_sec = timeout_ms / 1000,
@@ -127,9 +136,40 @@ static int firehose_read(struct qdl_device *qdl, int timeout_ms,
 	if (qdl->dev_type == QDL_DEVICE_SIM)
 		return 0;
 
-	do {
+	/*
+	 * The goal of firehose_read() is to find a response to a request among
+	 * one or more incoming messages AND to consume all incoming messages
+	 * (otherwise subsequent writes will time out).
+	 * The messages can be one of:
+	 * - <log/>
+	 * - <response value=""/>
+	 * - <response value="" rawmode="true"/>
+	 *
+	 * Generally <log/> messages are coming prior to the <response/>, but
+	 * on MSM8916 (at least) it's been observed that <log/> messages can
+	 * arrive after the <response/>.
+	 *
+	 * We therefor need to consume messages until there are no more
+	 * (timeout) and we have been able to parse out a response (using
+	 * @response_parser).
+	 *
+	 * In the special case that the <response/> contain an attribute
+	 * "rawmode=true", the device signals that it has entered a mode where
+	 * it will not send/receive XML-formatted commands. So, (at least for
+	 * reads) we need to shortcircuit the logic and directly terminate the
+	 * consumption of incoming data.
+	 */
+	for (;;) {
 		n = qdl_read(qdl, buf, sizeof(buf), 100);
-		if (n <= 0) {
+
+		/* Timeout after seeing a response, we're done waiting for logs */
+		if (n == -ETIMEDOUT && resp >= 0)
+			break;
+		/* We want to return resp on error, to not loose the reset resposne */
+		else if (n == -EIO)
+			break;
+
+		if (n == -ETIMEDOUT || n == 0) {
 			gettimeofday(&now, NULL);
 			if (timercmp(&now, &timeout, <))
 				continue;
@@ -144,11 +184,17 @@ static int firehose_read(struct qdl_device *qdl, int timeout_ms,
 		if (!node)
 			return error;
 
-		ret = response_parser(node, data);
+		ret = response_parser(node, data, &rawmode);
 		xmlFreeDoc(node->doc);
-	} while (ret == -EAGAIN);
 
-	return ret;
+		if (ret >= 0)
+			resp = ret;
+
+		if (rawmode)
+			break;
+	}
+
+	return resp;
 }
 
 static int firehose_write(struct qdl_device *qdl, xmlDoc *doc)
@@ -182,7 +228,7 @@ static int firehose_write(struct qdl_device *qdl, xmlDoc *doc)
 	for (;;) {
 		ux_debug("FIREHOSE WRITE: %s\n", s);
 		vip_gen_chunk_update(qdl, s, len);
-		ret = qdl_write(qdl, s, len);
+		ret = qdl_write(qdl, s, len, 1000);
 		saved_errno = errno;
 
 		/*
@@ -208,7 +254,8 @@ static int firehose_write(struct qdl_device *qdl, xmlDoc *doc)
  *
  * Return: max size supported by the remote, or negative errno on failure
  */
-static int firehose_configure_response_parser(xmlNode *node, void *data)
+static int firehose_configure_response_parser(xmlNode *node, void *data,
+					      bool *rawmode __unused)
 {
 	xmlChar *payload;
 	xmlChar *value;
@@ -253,45 +300,53 @@ static int firehose_configure_response_parser(xmlNode *node, void *data)
 }
 
 static int firehose_send_configure(struct qdl_device *qdl, size_t payload_size,
-				   bool skip_storage_init, const char *storage,
+				   bool skip_storage_init,
+				   enum qdl_storage_type storage,
 				   size_t *max_payload_size)
 {
+	static const char * const memory_names[] = {
+		[QDL_STORAGE_EMMC] = "emmc",
+		[QDL_STORAGE_NAND] = "nand",
+		[QDL_STORAGE_UFS] = "ufs",
+		[QDL_STORAGE_NVME] = "nvme",
+		[QDL_STORAGE_SPINOR] = "spinor",
+	};
 	xmlNode *root;
 	xmlNode *node;
 	xmlDoc *doc;
-	int ret;
 
 	doc = xmlNewDoc((xmlChar *)"1.0");
 	root = xmlNewNode(NULL, (xmlChar *)"data");
 	xmlDocSetRootElement(doc, root);
 
 	node = xmlNewChild(root, NULL, (xmlChar *)"configure", NULL);
-	xml_setpropf(node, "MemoryName", storage);
+	xml_setpropf(node, "MemoryName", memory_names[storage]);
 	xml_setpropf(node, "MaxPayloadSizeToTargetInBytes", "%lu", payload_size);
 	xml_setpropf(node, "verbose", "%d", 0);
 	xml_setpropf(node, "ZLPAwareHost", "%d", 1);
 	xml_setpropf(node, "SkipStorageInit", "%d", skip_storage_init);
 
-	ret = firehose_write(qdl, doc);
+	firehose_write(qdl, doc);
 	xmlFreeDoc(doc);
-	if (ret < 0)
-		return ret;
 
-	return firehose_read(qdl, 5000, firehose_configure_response_parser, max_payload_size);
+	return firehose_read(qdl, 100, firehose_configure_response_parser, max_payload_size);
 }
 
-static int firehose_configure(struct qdl_device *qdl, bool skip_storage_init,
-			      const char *storage)
+static int firehose_try_configure(struct qdl_device *qdl, bool skip_storage_init,
+				  enum qdl_storage_type storage)
 {
+	size_t max_sector_size;
+	size_t sector_sizes[] = { 512, 4096 };
+	struct read_op op;
 	size_t size = 0;
+	void *buf;
 	int ret;
+	unsigned int i;
 
 	ret = firehose_send_configure(qdl, qdl->max_payload_size, skip_storage_init,
 				      storage, &size);
-	if (ret < 0) {
-		ux_err("configure request failed\n");
-		return -1;
-	}
+	if (ret < 0)
+		return ret;
 
 	/*
 	 * In simulateion mode "remote" target can't propose different size, so
@@ -313,25 +368,58 @@ static int firehose_configure(struct qdl_device *qdl, bool skip_storage_init,
 
 	ux_debug("accepted max payload size: %zu\n", qdl->max_payload_size);
 
+	if (storage != QDL_STORAGE_NAND) {
+		max_sector_size = sector_sizes[ARRAY_SIZE(sector_sizes) - 1];
+		buf = malloc(max_sector_size);
+		memset(&op, 0, sizeof(op));
+		op.partition = 0;
+		op.start_sector = "1";
+		op.num_sectors = 1;
+
+		/*
+		 * Testing has shown that the loader will fail gracefully if a
+		 * read is issued with the wrong sector size, use this to attempt
+		 * to discover the storage device's sector size.
+		 */
+		for (i = 0; i < ARRAY_SIZE(sector_sizes); i++) {
+			op.sector_size = sector_sizes[i];
+
+			ret = firehose_read_buf(qdl, &op, buf, max_sector_size);
+			if (ret == 0) {
+				qdl->sector_size = sector_sizes[i];
+				break;
+			}
+		}
+	}
+
+	if (qdl->sector_size)
+		ux_debug("detected sector size of: %zd\n", qdl->sector_size);
+
 	return 0;
 }
 
 static int firehose_erase(struct qdl_device *qdl, struct program *program)
 {
+	unsigned int sector_size;
 	xmlNode *root;
 	xmlNode *node;
 	xmlDoc *doc;
 	int ret;
 
+	sector_size = program->sector_size ? : qdl->sector_size;
+
 	doc = xmlNewDoc((xmlChar *)"1.0");
 	root = xmlNewNode(NULL, (xmlChar *)"data");
 	xmlDocSetRootElement(doc, root);
 
 	node = xmlNewChild(root, NULL, (xmlChar *)"erase", NULL);
-	xml_setpropf(node, "SECTOR_SIZE_IN_BYTES", "%d", program->sector_size);
+	xml_setpropf(node, "SECTOR_SIZE_IN_BYTES", "%d", sector_size);
 	xml_setpropf(node, "num_partition_sectors", "%d", program->num_sectors);
 	xml_setpropf(node, "physical_partition_number", "%d", program->partition);
 	xml_setpropf(node, "start_sector", "%s", program->start_sector);
+	if (qdl->slot != UINT_MAX) {
+		xml_setpropf(node, "slot", "%u", qdl->slot);
+	}
 	if (program->is_nand) {
 		xml_setpropf(node, "PAGES_PER_BLOCK", "%d", program->pages_per_block);
 	}
@@ -356,6 +444,8 @@ out:
 static int firehose_program(struct qdl_device *qdl, struct program *program, int fd)
 {
 	unsigned int num_sectors;
+	unsigned int sector_size;
+	unsigned int zlp_timeout = 10000;
 	struct stat sb;
 	size_t chunk_size;
 	xmlNode *root;
@@ -364,25 +454,34 @@ static int firehose_program(struct qdl_device *qdl, struct program *program, int
 	void *buf;
 	time_t t0;
 	time_t t;
-	int left;
+	size_t left;
 	int ret;
 	int n;
+	size_t i;
 	uint32_t fill_value;
 
+	/*
+	 * ZLP has been measured to take up to 15 seconds on SPINOR devices,
+	 * let's double it to be on the safe side...
+	 */
+	if (qdl->storage_type == QDL_STORAGE_SPINOR)
+		zlp_timeout = 60000;
+
 	num_sectors = program->num_sectors;
+	sector_size = program->sector_size ? : qdl->sector_size;
 
 	ret = fstat(fd, &sb);
 	if (ret < 0)
 		err(1, "failed to stat \"%s\"\n", program->filename);
 
 	if (!program->sparse) {
-		num_sectors = (sb.st_size + program->sector_size - 1) / program->sector_size;
+		num_sectors = (sb.st_size + sector_size - 1) / sector_size;
 
 		if (program->num_sectors && num_sectors > program->num_sectors) {
 			ux_err("%s to big for %s truncated to %d\n",
 			       program->filename,
 			       program->label,
-			       program->num_sectors * program->sector_size);
+			       program->num_sectors * sector_size);
 			num_sectors = program->num_sectors;
 		}
 	}
@@ -396,10 +495,13 @@ static int firehose_program(struct qdl_device *qdl, struct program *program, int
 	xmlDocSetRootElement(doc, root);
 
 	node = xmlNewChild(root, NULL, (xmlChar *)"program", NULL);
-	xml_setpropf(node, "SECTOR_SIZE_IN_BYTES", "%d", program->sector_size);
+	xml_setpropf(node, "SECTOR_SIZE_IN_BYTES", "%d", sector_size);
 	xml_setpropf(node, "num_partition_sectors", "%d", num_sectors);
 	xml_setpropf(node, "physical_partition_number", "%d", program->partition);
 	xml_setpropf(node, "start_sector", "%s", program->start_sector);
+	if (qdl->slot != UINT_MAX) {
+		xml_setpropf(node, "slot", "%u", qdl->slot);
+	}
 	if (program->filename)
 		xml_setpropf(node, "filename", "%s", program->filename);
 
@@ -423,7 +525,7 @@ static int firehose_program(struct qdl_device *qdl, struct program *program, int
 	t0 = time(NULL);
 
 	if (!program->sparse) {
-		lseek(fd, (off_t)program->file_offset * program->sector_size, SEEK_SET);
+		lseek(fd, (off_t)program->file_offset * sector_size, SEEK_SET);
 	} else {
 		switch (program->sparse_chunk_type) {
 		case CHUNK_TYPE_RAW:
@@ -431,8 +533,8 @@ static int firehose_program(struct qdl_device *qdl, struct program *program, int
 			break;
 		case CHUNK_TYPE_FILL:
 			fill_value = program->sparse_fill_value;
-			for (n = 0; n < qdl->max_payload_size; n += sizeof(fill_value))
-				memcpy(buf + n, &fill_value, sizeof(fill_value));
+			for (i = 0; i < qdl->max_payload_size; i += sizeof(fill_value))
+				memcpy(buf + i, &fill_value, sizeof(fill_value));
 			break;
 		default:
 			ux_err("[SPARSE] invalid chunk type\n");
@@ -443,7 +545,7 @@ static int firehose_program(struct qdl_device *qdl, struct program *program, int
 	left = num_sectors;
 
 	ux_debug("FIREHOSE RAW BINARY WRITE: %s, %d bytes\n",
-		 program->filename, program->sector_size * num_sectors);
+		 program->filename, sector_size * num_sectors);
 
 	while (left > 0) {
 		/*
@@ -451,20 +553,20 @@ static int firehose_program(struct qdl_device *qdl, struct program *program, int
 		 * not for the whole binary.
 		 */
 		vip_gen_chunk_init(qdl);
-		chunk_size = MIN(qdl->max_payload_size / program->sector_size, left);
+		chunk_size = MIN(qdl->max_payload_size / sector_size, left);
 
 		if (!program->sparse || program->sparse_chunk_type != CHUNK_TYPE_FILL) {
-			n = read(fd, buf, chunk_size * program->sector_size);
+			n = read(fd, buf, chunk_size * sector_size);
 			if (n < 0) {
 				ux_err("failed to read %s\n", program->filename);
 				goto out;
 			}
 
-			if (n < qdl->max_payload_size)
+			if ((size_t)n < qdl->max_payload_size)
 				memset(buf + n, 0, qdl->max_payload_size - n);
 		}
 
-		vip_gen_chunk_update(qdl, buf, chunk_size * program->sector_size);
+		vip_gen_chunk_update(qdl, buf, chunk_size * sector_size);
 
 		ret = vip_transfer_handle_tables(qdl);
 		if (ret) {
@@ -482,7 +584,7 @@ static int firehose_program(struct qdl_device *qdl, struct program *program, int
 
 			vip_transfer_clear_status(qdl);
 		}
-		n = qdl_write(qdl, buf, chunk_size * program->sector_size);
+		n = qdl_write(qdl, buf, chunk_size * sector_size, zlp_timeout);
 		if (n < 0) {
 			ux_err("USB write failed for data chunk\n");
 			ret = firehose_read(qdl, 30000, firehose_generic_parser, NULL);
@@ -492,7 +594,7 @@ static int firehose_program(struct qdl_device *qdl, struct program *program, int
 			goto out;
 		}
 
-		if (n != chunk_size * program->sector_size) {
+		if ((size_t)n != chunk_size * sector_size) {
 			ux_err("USB write truncated\n");
 			ret = -1;
 			goto out;
@@ -512,7 +614,7 @@ static int firehose_program(struct qdl_device *qdl, struct program *program, int
 	} else if (t) {
 		ux_info("flashed \"%s\" successfully at %lukB/s\n",
 			program->label,
-			(unsigned long)program->sector_size * num_sectors / t / 1024);
+			(unsigned long)sector_size * num_sectors / t / 1024);
 	} else {
 		ux_info("flashed \"%s\" successfully\n",
 			program->label);
@@ -525,19 +627,21 @@ out:
 	return ret == FIREHOSE_ACK ? 0 : -1;
 }
 
-static int firehose_read_op(struct qdl_device *qdl, struct read_op *read_op, int fd)
+static int firehose_issue_read(struct qdl_device *qdl, struct read_op *read_op,
+			       int fd, void *out_buf, size_t out_len, bool quiet)
 {
+	unsigned int sector_size;
 	size_t chunk_size;
+	size_t out_offset = 0;
 	xmlNode *root;
 	xmlNode *node;
 	xmlDoc *doc;
 	void *buf;
 	time_t t0;
 	time_t t;
-	int left;
+	size_t left;
 	int ret;
 	int n;
-	bool expect_empty;
 
 	buf = malloc(qdl->max_payload_size);
 	if (!buf)
@@ -547,11 +651,16 @@ static int firehose_read_op(struct qdl_device *qdl, struct read_op *read_op, int
 	root = xmlNewNode(NULL, (xmlChar *)"data");
 	xmlDocSetRootElement(doc, root);
 
+	sector_size = read_op->sector_size ? : qdl->sector_size;
+
 	node = xmlNewChild(root, NULL, (xmlChar *)"read", NULL);
-	xml_setpropf(node, "SECTOR_SIZE_IN_BYTES", "%d", read_op->sector_size);
+	xml_setpropf(node, "SECTOR_SIZE_IN_BYTES", "%d", sector_size);
 	xml_setpropf(node, "num_partition_sectors", "%d", read_op->num_sectors);
 	xml_setpropf(node, "physical_partition_number", "%d", read_op->partition);
 	xml_setpropf(node, "start_sector", "%s", read_op->start_sector);
+	if (qdl->slot != UINT_MAX) {
+		xml_setpropf(node, "slot", "%u", qdl->slot);
+	}
 	if (read_op->filename)
 		xml_setpropf(node, "filename", "%s", read_op->filename);
 
@@ -563,43 +672,44 @@ static int firehose_read_op(struct qdl_device *qdl, struct read_op *read_op, int
 
 	ret = firehose_read(qdl, 10000, firehose_generic_parser, NULL);
 	if (ret) {
-		ux_err("failed to setup reading operation\n");
+		if (!quiet)
+			ux_err("failed to setup reading operation\n");
 		goto out;
 	}
 
 	t0 = time(NULL);
 
 	left = read_op->num_sectors;
-	expect_empty = false;
-	while (left > 0 || expect_empty) {
-		chunk_size = MIN(qdl->max_payload_size / read_op->sector_size, left);
+	while (left > 0) {
+		chunk_size = MIN(qdl->max_payload_size / sector_size, left);
 
-		n = qdl_read(qdl, buf, chunk_size * read_op->sector_size, 30000);
+		n = qdl_read(qdl, buf, chunk_size * sector_size, 30000);
 		if (n < 0) {
 			err(1, "failed to read");
 		}
 
-		if (n == 0 && expect_empty) {
-			// Every second transfer is empty during this stage.
-			expect_empty = false;
-			continue;
-		} else if (expect_empty) {
-			err(1, "expected empty transfer but received non-empty transfer during read");
-		} else if (n != chunk_size * read_op->sector_size) {
+		if ((size_t)n != chunk_size * sector_size) {
 			err(1, "failed to read full sector");
 		}
 
-		n = write(fd, buf, n);
+		if (out_buf) {
+			if ((size_t)n > out_len - out_offset)
+				n = out_len - out_offset;
 
-		if (n != chunk_size * read_op->sector_size) {
-			err(1, "failed to write");
+			memcpy(out_buf + out_offset, buf, n);
+			out_offset += n;
+		} else {
+			n = write(fd, buf, n);
+
+			if (n < 0 || (size_t)n != chunk_size * sector_size) {
+				err(1, "failed to write");
+			}
 		}
 
 		left -= chunk_size;
-#ifndef _WIN32
-		// on mac/linux, every other response is empty
-		expect_empty = true;
-#endif
+
+		if (!quiet)
+			ux_progress("%s", read_op->num_sectors - left, read_op->num_sectors, read_op->filename);
 	}
 
 	ret = firehose_read(qdl, 10000, firehose_generic_parser, NULL);
@@ -610,13 +720,15 @@ static int firehose_read_op(struct qdl_device *qdl, struct read_op *read_op, int
 
 	t = time(NULL) - t0;
 
-	if (t) {
-		ux_info("read \"%s\" successfully at %ldkB/s\n",
-			read_op->filename,
-			(unsigned long)read_op->sector_size * read_op->num_sectors / t / 1024);
-	} else {
-		ux_info("read \"%s\" successfully\n",
-			read_op->filename);
+	if (!quiet) {
+		if (t) {
+			ux_info("read \"%s\" successfully at %ldkB/s\n",
+				read_op->filename,
+				(unsigned long)sector_size * read_op->num_sectors / t / 1024);
+		} else {
+			ux_info("read \"%s\" successfully\n",
+				read_op->filename);
+		}
 	}
 
 out:
@@ -625,6 +737,16 @@ out:
 	return ret;
 }
 
+int firehose_read_buf(struct qdl_device *qdl, struct read_op *read_op, void *out_buf, size_t out_size)
+{
+	return firehose_issue_read(qdl, read_op, -1, out_buf, out_size, true);
+}
+
+static int firehose_read_op(struct qdl_device *qdl, struct read_op *read_op, int fd)
+{
+	return firehose_issue_read(qdl, read_op, fd, NULL, 0, false);
+}
+
 static int firehose_apply_patch(struct qdl_device *qdl, struct patch *patch)
 {
 	xmlNode *root;
@@ -646,6 +768,9 @@ static int firehose_apply_patch(struct qdl_device *qdl, struct patch *patch)
 	xml_setpropf(node, "size_in_bytes", "%d", patch->size_in_bytes);
 	xml_setpropf(node, "start_sector", "%s", patch->start_sector);
 	xml_setpropf(node, "value", "%s", patch->value);
+	if (qdl->slot != UINT_MAX) {
+		xml_setpropf(node, "slot", "%u", qdl->slot);
+	}
 
 	ret = firehose_write(qdl, doc);
 	if (ret < 0)
@@ -702,6 +827,9 @@ int firehose_apply_ufs_common(struct qdl_device *qdl, struct ufs_common *ufs)
 	xml_setpropf(node_to_send, "bInitActiveICCLevel", "%d", ufs->bInitActiveICCLevel);
 	xml_setpropf(node_to_send, "wPeriodicRTCUpdate", "%d", ufs->wPeriodicRTCUpdate);
 	xml_setpropf(node_to_send, "bConfigDescrLock", "%d", ufs->bConfigDescrLock);
+	if (qdl->slot != UINT_MAX) {
+		xml_setpropf(node_to_send, "slot", "%u", qdl->slot);
+	}
 
 	if (ufs->wb) {
 		xml_setpropf(node_to_send, "bWriteBoosterBufferPreserveUserSpaceEn",
@@ -734,6 +862,9 @@ int firehose_apply_ufs_body(struct qdl_device *qdl, struct ufs_body *ufs)
 	xml_setpropf(node_to_send, "bLogicalBlockSize", "%d", ufs->bLogicalBlockSize);
 	xml_setpropf(node_to_send, "bProvisioningType", "%d", ufs->bProvisioningType);
 	xml_setpropf(node_to_send, "wContextCapabilities", "%d", ufs->wContextCapabilities);
+	if (qdl->slot != UINT_MAX) {
+		xml_setpropf(node_to_send, "slot", "%u", qdl->slot);
+	}
 	if (ufs->desc)
 		xml_setpropf(node_to_send, "desc", "%s", ufs->desc);
 
@@ -754,6 +885,9 @@ int firehose_apply_ufs_epilogue(struct qdl_device *qdl, struct ufs_epilogue *ufs
 
 	xml_setpropf(node_to_send, "LUNtoGrow", "%d", ufs->LUNtoGrow);
 	xml_setpropf(node_to_send, "commit", "%d", commit);
+	if (qdl->slot != UINT_MAX) {
+		xml_setpropf(node_to_send, "slot", "%u", qdl->slot);
+	}
 
 	ret = firehose_send_single_tag(qdl, node_to_send);
 	if (ret)
@@ -820,8 +954,60 @@ static int firehose_reset(struct qdl_device *qdl)
 	return ret == FIREHOSE_ACK ? 0 : -1;
 }
 
-int firehose_run(struct qdl_device *qdl, const char *incdir,
-		 const char *storage, bool allow_missing)
+static int firehose_detect_and_configure(struct qdl_device *qdl,
+					 bool skip_storage_init __unused,
+					 enum qdl_storage_type storage,
+					 unsigned int timeout_s)
+{
+	struct timeval timeout = { .tv_sec = timeout_s };
+	struct timeval now;
+	int ret;
+
+	gettimeofday(&now, NULL);
+	timeradd(&now, &timeout, &timeout);
+	for (;;) {
+		ret = firehose_try_configure(qdl, false, storage);
+
+		if (ret == FIREHOSE_ACK) {
+			break;
+		} else if (ret != -ETIMEDOUT) {
+			ux_err("configure request failed\n");
+			return -1;
+		}
+
+		gettimeofday(&now, NULL);
+		if (timercmp(&now, &timeout, >)) {
+			ux_err("failed to detect firehose programmer\n");
+			return -1;
+		}
+	}
+
+	return 0;
+}
+
+int firehose_provision(struct qdl_device *qdl)
+{
+	int ret;
+
+	ret = firehose_detect_and_configure(qdl, true, QDL_STORAGE_UFS, 5);
+	if (ret)
+		return ret;
+
+	ret = ufs_provisioning_execute(qdl, firehose_apply_ufs_common,
+				       firehose_apply_ufs_body,
+				       firehose_apply_ufs_epilogue);
+	if (!ret)
+		ux_info("UFS provisioning succeeded\n");
+	else
+		ux_info("UFS provisioning failed\n");
+
+	firehose_reset(qdl);
+
+	return ret;
+
+}
+
+int firehose_run(struct qdl_device *qdl)
 {
 	bool multiple;
 	int bootable;
@@ -829,26 +1015,15 @@ int firehose_run(struct qdl_device *qdl, const char *incdir,
 
 	ux_info("waiting for programmer...\n");
 
-	firehose_read(qdl, 5000, firehose_generic_parser, NULL);
-
-	if (ufs_need_provisioning()) {
-		ret = firehose_configure(qdl, true, storage);
-		if (ret)
-			return ret;
-		ret = ufs_provisioning_execute(qdl, firehose_apply_ufs_common,
-					       firehose_apply_ufs_body,
-					       firehose_apply_ufs_epilogue);
-		if (!ret)
-			ux_info("UFS provisioning succeeded\n");
-		else
-			ux_info("UFS provisioning failed\n");
-
-		firehose_reset(qdl);
-
+	ret = firehose_detect_and_configure(qdl, true, qdl->storage_type, 5);
+	if (ret)
 		return ret;
-	}
 
-	ret = firehose_configure(qdl, false, storage);
+	ret = read_resolve_gpt_deferrals(qdl);
+	if (ret)
+		return ret;
+
+	ret = program_resolve_gpt_deferrals(qdl);
 	if (ret)
 		return ret;
 
@@ -856,7 +1031,7 @@ int firehose_run(struct qdl_device *qdl, const char *incdir,
 	if (ret)
 		return ret;
 
-	ret = program_execute(qdl, firehose_program, incdir, allow_missing);
+	ret = program_execute(qdl, firehose_program);
 	if (ret)
 		return ret;
 
@@ -864,7 +1039,7 @@ int firehose_run(struct qdl_device *qdl, const char *incdir,
 	if (ret)
 		return ret;
 
-	ret = read_op_execute(qdl, firehose_read_op, incdir);
+	ret = read_op_execute(qdl, firehose_read_op);
 	if (ret)
 		return ret;
 
