@@ -13,8 +13,8 @@
 #include "patch.h"
 #include "qdl.h"
 
-static struct patch *patches;
-static struct patch *patches_last;
+static struct list_head patches = LIST_INIT(patches);
+static bool patches_loaded;
 
 int patch_load(const char *patch_file)
 {
@@ -59,17 +59,13 @@ int patch_load(const char *patch_file)
 			continue;
 		}
 
-		if (patches) {
-			patches_last->next = patch;
-			patches_last = patch;
-		} else {
-			patches = patch;
-			patches_last = patch;
-		}
+		list_add(&patches, &patch->node);
 	}
 
 	xmlFreeDoc(doc);
 
+	patches_loaded = true;
+
 	return 0;
 }
 
@@ -80,12 +76,21 @@ int patch_execute(struct qdl_device *qdl, int (*apply)(struct qdl_device *qdl, s
 	unsigned int idx = 0;
 	int ret;
 
-	for (patch = patches; patch; patch = patch->next) {
+	if (!patches_loaded)
+		return 0;
+
+	list_for_each_entry(patch, &patches, node) {
+		if (!patch->filename)
+			continue;
+
 		if (!strcmp(patch->filename, "DISK"))
 			count++;
 	}
 
-	for (patch = patches; patch; patch = patch->next) {
+	list_for_each_entry(patch, &patches, node) {
+		if (!patch->filename)
+			continue;
+
 		if (strcmp(patch->filename, "DISK"))
 			continue;
 
@@ -103,11 +108,10 @@ int patch_execute(struct qdl_device *qdl, int (*apply)(struct qdl_device *qdl, s
 
 void free_patches(void)
 {
-	struct patch *patch = patches;
+	struct patch *patch;
 	struct patch *next;
 
-	for (patch = patches; patch; patch = next) {
-		next = patch->next;
+	list_for_each_entry_safe(patch, next, &patches, node) {
 		free((void *)patch->filename);
 		free((void *)patch->start_sector);
 		free((void *)patch->value);
@@ -115,5 +119,5 @@ void free_patches(void)
 		free(patch);
 	}
 
-	patches = NULL;
+	list_init(&patches);
 }
