@@ -18,9 +18,9 @@
 #include "qdl.h"
 #include "oscompat.h"
 #include "sparse.h"
+#include "gpt.h"
 
-static struct program *programes;
-static struct program *programes_last;
+static struct list_head programs = LIST_INIT(programs);
 
 static int load_erase_tag(xmlNode *node, bool is_nand)
 {
@@ -46,22 +46,14 @@ static int load_erase_tag(xmlNode *node, bool is_nand)
 		return -EINVAL;
 	}
 
-	if (programes) {
-		programes_last->next = program;
-		programes_last = program;
-	} else {
-		programes = program;
-		programes_last = program;
-	}
+	list_add(&programs, &program->node);
 
 	return 0;
 }
 
-static struct program *program_load_sparse(struct program *program, int fd)
+static int program_load_sparse(struct program *program, int fd)
 {
 	struct program *program_sparse = NULL;
-	struct program *programes_sparse = NULL;
-	struct program *programes_sparse_last = NULL;
 	char tmp[PATH_MAX];
 
 	sparse_header_t sparse_header;
@@ -77,20 +69,19 @@ static struct program *program_load_sparse(struct program *program, int fd)
 		 * for a partition node but lacks a sparse header,
 		 * it will be validated against the defined partition size.
 		 * If the sizes match, it is likely that the 'sparse="true"' attribute
-		 * was set by mistake.
+		 * was set by mistake, fix the sparse flag and add the
+		 * program entry to the list.
 		 */
 		if ((off_t)program->sector_size * program->num_sectors ==
 		    lseek(fd, 0, SEEK_END)) {
-			program_sparse = calloc(1, sizeof(struct program));
-			memcpy(program_sparse, program, sizeof(struct program));
-			program_sparse->sparse = false;
-			program_sparse->next = NULL;
-			return program_sparse;
+			program->sparse = false;
+			list_add(&programs, &program->node);
+			return 0;
 		}
 
 		ux_err("[PROGRAM] Unable to parse sparse header at %s...failed\n",
 		       program->filename);
-		return NULL;
+		return -1;
 	}
 
 	for (uint32_t i = 0; i < sparse_header.total_chunks; ++i) {
@@ -101,7 +92,7 @@ static struct program *program_load_sparse(struct program *program, int fd)
 		if (chunk_type < 0) {
 			ux_err("[PROGRAM] Unable to parse sparse chunk %i at %s...failed\n",
 			       i, program->filename);
-			return NULL;
+			return -1;
 		}
 
 		if (chunk_size == 0)
@@ -110,7 +101,7 @@ static struct program *program_load_sparse(struct program *program, int fd)
 		if (chunk_size % program->sector_size != 0) {
 			ux_err("[SPARSE] File chunk #%u size %" PRIu64 " is not a sector-multiple\n",
 			       i, chunk_size);
-			return NULL;
+			return -1;
 		}
 
 		if (chunk_size / program->sector_size >= UINT_MAX) {
@@ -120,51 +111,32 @@ static struct program *program_load_sparse(struct program *program, int fd)
 			 */
 			ux_err("[SPARSE] File chunk #%u size %" PRIu64 " is too large\n",
 			       i, chunk_size);
-			return NULL;
+			return -1;
 		}
 
-		switch (chunk_type) {
-		case CHUNK_TYPE_RAW:
+		if (chunk_type == CHUNK_TYPE_RAW || chunk_type == CHUNK_TYPE_FILL) {
 			program_sparse = calloc(1, sizeof(struct program));
-			memcpy(program_sparse, program, sizeof(struct program));
+			program_sparse->pages_per_block = program->pages_per_block;
+			program_sparse->sector_size = program->sector_size;
+			program_sparse->file_offset = program->file_offset;
+			program_sparse->filename = strdup(program->filename);
+			program_sparse->label = strdup(program->label);
+			program_sparse->partition = program->partition;
+			program_sparse->sparse = program->sparse;
+			program_sparse->start_sector = strdup(program->start_sector);
+			program_sparse->last_sector = program->last_sector;
+			program_sparse->is_nand = program->is_nand;
+			program_sparse->is_erase = false;
 
-			program_sparse->next = NULL;
+			program_sparse->sparse_chunk_type = chunk_type;
 			program_sparse->num_sectors = chunk_size / program->sector_size;
 
-			program_sparse->sparse_chunk_type = CHUNK_TYPE_RAW;
-			program_sparse->sparse_offset = sparse_offset;
+			if (chunk_type == CHUNK_TYPE_RAW)
+				program_sparse->sparse_offset = sparse_offset;
+			else
+				program_sparse->sparse_fill_value = sparse_fill_value;
 
-			if (programes_sparse) {
-				programes_sparse_last->next = program_sparse;
-				programes_sparse_last = program_sparse;
-			} else {
-				programes_sparse = program_sparse;
-				programes_sparse_last = program_sparse;
-			}
-
-			break;
-		case CHUNK_TYPE_FILL:
-
-			program_sparse = calloc(1, sizeof(struct program));
-			memcpy(program_sparse, program, sizeof(struct program));
-
-			program_sparse->next = NULL;
-			program_sparse->num_sectors = chunk_size / program->sector_size;
-
-			program_sparse->sparse_chunk_type = CHUNK_TYPE_FILL;
-			program_sparse->sparse_fill_value = sparse_fill_value;
-
-			if (programes_sparse) {
-				programes_sparse_last->next = program_sparse;
-				programes_sparse_last = program_sparse;
-			} else {
-				programes_sparse = program_sparse;
-				programes_sparse_last = program_sparse;
-			}
-
-			break;
-		default:
-			break;
+			list_add(&programs, &program_sparse->node);
 		}
 
 		start_sector = (unsigned int)strtoul(program->start_sector, NULL, 0);
@@ -173,13 +145,16 @@ static struct program *program_load_sparse(struct program *program, int fd)
 		program->start_sector = strdup(tmp);
 	}
 
-	return programes_sparse;
+	return 0;
 }
 
-static int load_program_tag(xmlNode *node, bool is_nand)
+static int load_program_tag(xmlNode *node, bool is_nand, bool allow_missing, const char *incdir)
 {
 	struct program *program;
+	char tmp[PATH_MAX];
 	int errors = 0;
+	int ret;
+	int fd = -1;
 
 	program = calloc(1, sizeof(struct program));
 
@@ -208,18 +183,51 @@ static int load_program_tag(xmlNode *node, bool is_nand)
 		return -EINVAL;
 	}
 
-	if (programes) {
-		programes_last->next = program;
-		programes_last = program;
-	} else {
-		programes = program;
-		programes_last = program;
+	if (program->filename) {
+		if (incdir) {
+			snprintf(tmp, PATH_MAX, "%s/%s", incdir, program->filename);
+			if (access(tmp, F_OK) != -1) {
+				free((void *)program->filename);
+				program->filename = strdup(tmp);
+			}
+		}
+
+		fd = open(program->filename, O_RDONLY | O_BINARY);
+		if (fd < 0) {
+			ux_info("unable to open %s", program->filename);
+			if (!allow_missing) {
+				ux_info("...failing\n");
+				return -1;
+			}
+			ux_info("...ignoring\n");
+
+			free((void *)program->filename);
+			program->filename = NULL;
+		}
 	}
 
+	if (fd >= 0 && program->sparse) {
+		ret = program_load_sparse(program, fd);
+		if (ret < 0)
+			return -1;
+
+		/*
+		 * Chunks were added to the program list, drop the filename of
+		 * the parent, to prevent this from being written to the device
+		 */
+		free((void *)program->filename);
+		program->filename = NULL;
+	}
+
+	list_add(&programs, &program->node);
+
+	if (fd >= 0)
+		close(fd);
+
 	return 0;
 }
 
-int program_load(const char *program_file, bool is_nand)
+int program_load(const char *program_file, bool is_nand, bool allow_missing, const char *incdir)
 {
 	xmlNode *node;
 	xmlNode *root;
@@ -240,7 +248,7 @@ int program_load(const char *program_file, bool is_nand)
 		if (!xmlStrcmp(node->name, (xmlChar *)"erase"))
 			errors = load_erase_tag(node, is_nand);
 		else if (!xmlStrcmp(node->name, (xmlChar *)"program"))
-			errors = load_program_tag(node, is_nand);
+			errors = load_program_tag(node, is_nand, allow_missing, incdir);
 		else {
 			ux_err("unrecognized tag \"%s\" in program-type file \"%s\"\n", node->name, program_file);
 			errors = -EINVAL;
@@ -256,55 +264,24 @@ out:
 	return errors;
 }
 
-int program_execute(struct qdl_device *qdl, int (*apply)(struct qdl_device *qdl, struct program *program, int fd),
-		    const char *incdir, bool allow_missing)
+int program_execute(struct qdl_device *qdl, int (*apply)(struct qdl_device *qdl, struct program *program, int fd))
 {
 	struct program *program;
-	struct program *program_sparse;
-	const char *filename;
-	char tmp[PATH_MAX];
 	int ret;
 	int fd;
 
-	for (program = programes; program; program = program->next) {
+	list_for_each_entry(program, &programs, node) {
 		if (program->is_erase || !program->filename)
 			continue;
 
-		filename = program->filename;
-		if (incdir) {
-			snprintf(tmp, PATH_MAX, "%s/%s", incdir, filename);
-			if (access(tmp, F_OK) != -1)
-				filename = tmp;
-		}
-
-		fd = open(filename, O_RDONLY | O_BINARY);
+		fd = open(program->filename, O_RDONLY | O_BINARY);
 
 		if (fd < 0) {
-			ux_info("unable to open %s", program->filename);
-			if (!allow_missing) {
-				ux_info("...failing\n");
-				return -1;
-			}
-			ux_info("...ignoring\n");
-			continue;
-		}
-
-		if (!program->sparse) {
-			ret = apply(qdl, program, fd);
-		} else {
-			program_sparse = program_load_sparse(program, fd);
-			if (!program_sparse) {
-				ux_err("[PROGRAM] load sparse failed\n");
-				return -EINVAL;
-			}
-
-			for (; program_sparse; program_sparse = program_sparse->next) {
-				ret = apply(qdl, program_sparse, fd);
-				if (ret)
-					break;
-			}
+			ux_err("unable to open %s\n", program->filename);
+			return -1;
 		}
 
+		ret = apply(qdl, program, fd);
 		close(fd);
 		if (ret)
 			return ret;
@@ -318,7 +295,7 @@ int erase_execute(struct qdl_device *qdl, int (*apply)(struct qdl_device *qdl, s
 	struct program *program;
 	int ret;
 
-	for (program = programes; program; program = program->next) {
+	list_for_each_entry(program, &programs, node) {
 		if (!program->is_erase)
 			continue;
 
@@ -335,7 +312,7 @@ static struct program *program_find_partition(const char *partition)
 	struct program *program;
 	const char *label;
 
-	for (program = programes; program; program = program->next) {
+	list_for_each_entry(program, &programs, node) {
 		label = program->label;
 		if (!label)
 			continue;
@@ -409,16 +386,74 @@ int program_is_sec_partition_flashed(void)
 
 void free_programs(void)
 {
-	struct program *program = programes;
+	struct program *program;
 	struct program *next;
 
-	for (program = programes; program; program = next) {
-		next = program->next;
+	list_for_each_entry_safe(program, next, &programs, node) {
 		free((void *)program->filename);
 		free((void *)program->label);
 		free((void *)program->start_sector);
+		free((void *)program->gpt_partition);
 		free(program);
 	}
 
-	programes = NULL;
+	list_init(&programs);
+}
+
+int program_cmd_add(const char *address, const char *filename)
+{
+	unsigned int start_sector;
+	unsigned int num_sectors;
+	struct program *program;
+	char *gpt_partition;
+	int partition;
+	char buf[20];
+	int ret;
+
+	ret = parse_storage_address(address, &partition, &start_sector, &num_sectors, &gpt_partition);
+	if (ret < 0)
+		return ret;
+
+	program = calloc(1, sizeof(struct program));
+
+	program->sector_size = 0;
+	program->file_offset = 0;
+	program->filename = filename ? strdup(filename) : NULL;
+	program->label = filename ? strdup(filename) : NULL;
+	program->num_sectors = num_sectors;
+	program->partition = partition;
+	program->sparse = false;
+	sprintf(buf, "%u", start_sector);
+	program->start_sector = strdup(buf);
+	program->last_sector = 0;
+	program->is_nand = false;
+	program->is_erase = false;
+	program->gpt_partition = gpt_partition;
+
+	list_add(&programs, &program->node);
+
+	return 0;
+}
+
+int program_resolve_gpt_deferrals(struct qdl_device *qdl)
+{
+	struct program *program;
+	unsigned int start_sector;
+	char buf[20];
+	int ret;
+
+	list_for_each_entry(program, &programs, node) {
+		if (!program->gpt_partition)
+			continue;
+
+		ret = gpt_find_by_name(qdl, program->gpt_partition, &program->partition,
+				       &start_sector, &program->num_sectors);
+		if (ret < 0)
+			return -1;
+
+		sprintf(buf, "%u", start_sector);
+		program->start_sector = strdup(buf);
+	}
+
+	return 0;
 }
