@@ -5,7 +5,7 @@
 #include <sys/types.h>
 #include <stdbool.h>
 #include <stdint.h>
-#include "qdl.h"
+#include "list.h"
 
 struct program {
 	unsigned int pages_per_block;
@@ -14,7 +14,7 @@ struct program {
 	const char *filename;
 	const char *label;
 	unsigned int num_sectors;
-	unsigned int partition;
+	int partition;
 	bool sparse;
 	const char *start_sector;
 	unsigned int last_sector;
@@ -26,15 +26,20 @@ struct program {
 	uint32_t sparse_fill_value;
 	off_t sparse_offset;
 
-	struct program *next;
+	const char *gpt_partition;
+
+	struct list_head node;
 };
 
-int program_load(const char *program_file, bool is_nand);
-int program_execute(struct qdl_device *qdl, int (*apply)(struct qdl_device *qdl, struct program *program, int fd),
-		    const char *incdir, bool allow_missing);
+struct qdl_device;
+
+int program_load(const char *program_file, bool is_nand, bool allow_missing, const char *incdir);
+int program_execute(struct qdl_device *qdl, int (*apply)(struct qdl_device *qdl, struct program *program, int fd));
 int erase_execute(struct qdl_device *qdl, int (*apply)(struct qdl_device *qdl, struct program *program));
 int program_find_bootable_partition(bool *multiple_found);
 int program_is_sec_partition_flashed(void);
+int program_cmd_add(const char *address, const char *filename);
+int program_resolve_gpt_deferrals(struct qdl_device *qdl);
 
 void free_programs(void);
 
